@@ -1,18 +1,15 @@
-from inspect import isclass, signature, Signature
+from copy import copy
+from inspect import isclass, signature, Signature, getmodule
 from typing import (
     Annotated,
     AnyStr,
     Callable,
-    Dict,
     Literal,
     NamedTuple,
     NewType,
     Optional,
     Protocol,
-    Set,
     Sequence,
-    Tuple,
-    Type,
     TypeGuard,
     Union,
     get_args,
@@ -22,19 +19,19 @@ from typing import (
 import ast
 import builtins
 import collections
+import dataclasses
 import operator
 import sys
+import typing
+import warnings
 from functools import cached_property
 from dataclasses import dataclass, field
-from types import MethodDescriptorType, ModuleType
+from types import MethodDescriptorType, ModuleType, MethodType
 
 from IPython.utils.decorators import undoc
 
 
-if sys.version_info < (3, 11):
-    from typing_extensions import Self, LiteralString
-else:
-    from typing import Self, LiteralString
+from typing import Self, LiteralString
 
 if sys.version_info < (3, 12):
     from typing_extensions import TypeAliasType
@@ -99,7 +96,8 @@ class EvaluationPolicy:
     allow_builtins_access: bool = False
     allow_all_operations: bool = False
     allow_any_calls: bool = False
-    allowed_calls: Set[Callable] = field(default_factory=set)
+    allow_auto_import: bool = False
+    allowed_calls: set[Callable] = field(default_factory=set)
 
     def can_get_item(self, value, item):
         return self.allow_item_access
@@ -107,7 +105,7 @@ class EvaluationPolicy:
     def can_get_attr(self, value, attr):
         return self.allow_attr_access
 
-    def can_operate(self, dunders: Tuple[str, ...], a, b=None):
+    def can_operate(self, dunders: tuple[str, ...], a, b=None):
         if self.allow_all_operations:
             return True
 
@@ -127,14 +125,24 @@ class EvaluationPolicy:
 def _get_external(module_name: str, access_path: Sequence[str]):
     """Get value from external module given a dotted access path.
 
+    Only gets value if the module is already imported.
+
     Raises:
     * `KeyError` if module is removed not found, and
     * `AttributeError` if access path does not match an exported object
     """
-    member_type = sys.modules[module_name]
-    for attr in access_path:
-        member_type = getattr(member_type, attr)
-    return member_type
+    try:
+        member_type = sys.modules[module_name]
+        # standard module
+        for attr in access_path:
+            member_type = getattr(member_type, attr)
+        return member_type
+    except (KeyError, AttributeError):
+        # handle modules in namespace packages
+        module_path = ".".join([module_name, *access_path])
+        if module_path in sys.modules:
+            return sys.modules[module_path]
+        raise
 
 
 def _has_original_dunder_external(
@@ -144,18 +152,26 @@ def _has_original_dunder_external(
     method_name: str,
 ):
     if module_name not in sys.modules:
-        # LBYLB as it is faster
-        return False
+        full_module_path = ".".join([module_name, *access_path])
+        if full_module_path not in sys.modules:
+            # LBYLB as it is faster
+            return False
     try:
         member_type = _get_external(module_name, access_path)
         value_type = type(value)
         if type(value) == member_type:
             return True
+        if isinstance(member_type, ModuleType):
+            value_module = getmodule(value_type)
+            if not value_module or not value_module.__name__:
+                return False
+            if value_module.__name__.startswith(member_type.__name__):
+                return True
         if method_name == "__getattribute__":
             # we have to short-circuit here due to an unresolved issue in
             # `isinstance` implementation: https://bugs.python.org/issue32683
             return False
-        if isinstance(value, member_type):
+        if not isinstance(member_type, ModuleType) and isinstance(value, member_type):
             method = getattr(value_type, method_name, None)
             member_method = getattr(member_type, method_name, None)
             if member_method == method:
@@ -190,35 +206,49 @@ def _has_original_dunder(
     return False
 
 
+def _coerce_path_to_tuples(
+    allow_list: set[tuple[str, ...] | str],
+) -> set[tuple[str, ...]]:
+    """Replace dotted paths on the provided allow-list with tuples."""
+    return {
+        path if isinstance(path, tuple) else tuple(path.split("."))
+        for path in allow_list
+    }
+
+
 @undoc
 @dataclass
 class SelectivePolicy(EvaluationPolicy):
-    allowed_getitem: Set[InstancesHaveGetItem] = field(default_factory=set)
-    allowed_getitem_external: Set[Tuple[str, ...]] = field(default_factory=set)
+    allowed_getitem: set[InstancesHaveGetItem] = field(default_factory=set)
+    allowed_getitem_external: set[tuple[str, ...] | str] = field(default_factory=set)
 
-    allowed_getattr: Set[MayHaveGetattr] = field(default_factory=set)
-    allowed_getattr_external: Set[Tuple[str, ...]] = field(default_factory=set)
+    allowed_getattr: set[MayHaveGetattr] = field(default_factory=set)
+    allowed_getattr_external: set[tuple[str, ...] | str] = field(default_factory=set)
 
-    allowed_operations: Set = field(default_factory=set)
-    allowed_operations_external: Set[Tuple[str, ...]] = field(default_factory=set)
+    allowed_operations: set = field(default_factory=set)
+    allowed_operations_external: set[tuple[str, ...] | str] = field(default_factory=set)
 
-    _operation_methods_cache: Dict[str, Set[Callable]] = field(
+    allow_getitem_on_types: bool = field(default_factory=bool)
+
+    _operation_methods_cache: dict[str, set[Callable]] = field(
         default_factory=dict, init=False
     )
 
     def can_get_attr(self, value, attr):
+        allowed_getattr_external = _coerce_path_to_tuples(self.allowed_getattr_external)
+
         has_original_attribute = _has_original_dunder(
             value,
             allowed_types=self.allowed_getattr,
             allowed_methods=self._getattribute_methods,
-            allowed_external=self.allowed_getattr_external,
+            allowed_external=allowed_getattr_external,
             method_name="__getattribute__",
         )
         has_original_attr = _has_original_dunder(
             value,
             allowed_types=self.allowed_getattr,
             allowed_methods=self._getattr_methods,
-            allowed_external=self.allowed_getattr_external,
+            allowed_external=allowed_getattr_external,
             method_name="__getattr__",
         )
 
@@ -250,7 +280,7 @@ class SelectivePolicy(EvaluationPolicy):
                 return True  # pragma: no cover
 
             # Properties in subclasses of allowed types may be ok if not changed
-            for module_name, *access_path in self.allowed_getattr_external:
+            for module_name, *access_path in allowed_getattr_external:
                 try:
                     external_class = _get_external(module_name, access_path)
                     external_class_attr_val = getattr(external_class, attr)
@@ -262,15 +292,26 @@ class SelectivePolicy(EvaluationPolicy):
 
     def can_get_item(self, value, item):
         """Allow accessing `__getiitem__` of allow-listed instances unless it was not modified."""
+        allowed_getitem_external = _coerce_path_to_tuples(self.allowed_getitem_external)
+        if self.allow_getitem_on_types:
+            # e.g. Union[str, int] or Literal[True, 1]
+            if isinstance(value, (typing._SpecialForm, typing._BaseGenericAlias)):
+                return True
+            # PEP 560 e.g. list[str]
+            if isinstance(value, type) and hasattr(value, "__class_getitem__"):
+                return True
         return _has_original_dunder(
             value,
             allowed_types=self.allowed_getitem,
             allowed_methods=self._getitem_methods,
-            allowed_external=self.allowed_getitem_external,
+            allowed_external=allowed_getitem_external,
             method_name="__getitem__",
         )
 
-    def can_operate(self, dunders: Tuple[str, ...], a, b=None):
+    def can_operate(self, dunders: tuple[str, ...], a, b=None):
+        allowed_operations_external = _coerce_path_to_tuples(
+            self.allowed_operations_external
+        )
         objects = [a]
         if b is not None:
             objects.append(b)
@@ -280,7 +321,7 @@ class SelectivePolicy(EvaluationPolicy):
                     obj,
                     allowed_types=self.allowed_operations,
                     allowed_methods=self._operator_dunder_methods(dunder),
-                    allowed_external=self.allowed_operations_external,
+                    allowed_external=allowed_operations_external,
                     method_name=dunder,
                 )
                 for dunder in dunders
@@ -288,7 +329,7 @@ class SelectivePolicy(EvaluationPolicy):
             ]
         )
 
-    def _operator_dunder_methods(self, dunder: str) -> Set[Callable]:
+    def _operator_dunder_methods(self, dunder: str) -> set[Callable]:
         if dunder not in self._operation_methods_cache:
             self._operation_methods_cache[dunder] = self._safe_get_methods(
                 self.allowed_operations, dunder
@@ -296,18 +337,18 @@ class SelectivePolicy(EvaluationPolicy):
         return self._operation_methods_cache[dunder]
 
     @cached_property
-    def _getitem_methods(self) -> Set[Callable]:
+    def _getitem_methods(self) -> set[Callable]:
         return self._safe_get_methods(self.allowed_getitem, "__getitem__")
 
     @cached_property
-    def _getattr_methods(self) -> Set[Callable]:
+    def _getattr_methods(self) -> set[Callable]:
         return self._safe_get_methods(self.allowed_getattr, "__getattr__")
 
     @cached_property
-    def _getattribute_methods(self) -> Set[Callable]:
+    def _getattribute_methods(self) -> set[Callable]:
         return self._safe_get_methods(self.allowed_getattr, "__getattribute__")
 
-    def _safe_get_methods(self, classes, name) -> Set[Callable]:
+    def _safe_get_methods(self, classes, name) -> set[Callable]:
         return {
             method
             for class_ in classes
@@ -320,18 +361,30 @@ class _DummyNamedTuple(NamedTuple):
     """Used internally to retrieve methods of named tuple instance."""
 
 
-class EvaluationContext(NamedTuple):
+EvaluationPolicyName = Literal["forbidden", "minimal", "limited", "unsafe", "dangerous"]
+
+
+@dataclass
+class EvaluationContext:
     #: Local namespace
     locals: dict
     #: Global namespace
     globals: dict
     #: Evaluation policy identifier
-    evaluation: Literal["forbidden", "minimal", "limited", "unsafe", "dangerous"] = (
-        "forbidden"
-    )
+    evaluation: EvaluationPolicyName = "forbidden"
     #: Whether the evaluation of code takes place inside of a subscript.
     #: Useful for evaluating ``:-1, 'col'`` in ``df[:-1, 'col']``.
     in_subscript: bool = False
+    #: Auto import method
+    auto_import: Callable[list[str], ModuleType] | None = None
+    #: Overrides for evaluation policy
+    policy_overrides: dict = field(default_factory=dict)
+    #: Transient local namespace used to store mocks
+    transient_locals: dict = field(default_factory=dict)
+
+    def replace(self, /, **changes):
+        """Return a new copy of the context, with specified changes"""
+        return dataclasses.replace(self, **changes)
 
 
 class _IdentitySubscript:
@@ -379,17 +432,17 @@ def guarded_eval(code: str, context: EvaluationContext):
         locals_ = locals_.copy()
         locals_[SUBSCRIPT_MARKER] = IDENTITY_SUBSCRIPT
         code = SUBSCRIPT_MARKER + "[" + code + "]"
-        context = EvaluationContext(**{**context._asdict(), **{"locals": locals_}})
+        context = context.replace(locals=locals_)
 
     if context.evaluation == "dangerous":
         return eval(code, context.globals, context.locals)
 
-    expression = ast.parse(code, mode="eval")
+    node = ast.parse(code, mode="exec")
 
-    return eval_node(expression, context)
+    return eval_node(node, context)
 
 
-BINARY_OP_DUNDERS: Dict[Type[ast.operator], Tuple[str]] = {
+BINARY_OP_DUNDERS: dict[type[ast.operator], tuple[str]] = {
     ast.Add: ("__add__",),
     ast.Sub: ("__sub__",),
     ast.Mult: ("__mul__",),
@@ -405,7 +458,7 @@ BINARY_OP_DUNDERS: Dict[Type[ast.operator], Tuple[str]] = {
     ast.MatMult: ("__matmul__",),
 }
 
-COMP_OP_DUNDERS: Dict[Type[ast.cmpop], Tuple[str, ...]] = {
+COMP_OP_DUNDERS: dict[type[ast.cmpop], tuple[str, ...]] = {
     ast.Eq: ("__eq__",),
     ast.NotEq: ("__ne__", "__eq__"),
     ast.Lt: ("__lt__", "__gt__"),
@@ -416,7 +469,7 @@ COMP_OP_DUNDERS: Dict[Type[ast.cmpop], Tuple[str, ...]] = {
     # Note: ast.Is, ast.IsNot, ast.NotIn are handled specially
 }
 
-UNARY_OP_DUNDERS: Dict[Type[ast.unaryop], Tuple[str, ...]] = {
+UNARY_OP_DUNDERS: dict[type[ast.unaryop], tuple[str, ...]] = {
     ast.USub: ("__neg__",),
     ast.UAdd: ("__pos__",),
     # we have to check both __inv__ and __invert__!
@@ -435,8 +488,8 @@ class _Duck:
     """A dummy class used to create objects pretending to have given attributes"""
 
     def __init__(self, attributes: Optional[dict] = None, items: Optional[dict] = None):
-        self.attributes = attributes or {}
-        self.items = items or {}
+        self.attributes = attributes if attributes is not None else {}
+        self.items = items if items is not None else {}
 
     def __getattr__(self, attr: str):
         return self.attributes[attr]
@@ -457,7 +510,7 @@ class _Duck:
         return self.items.keys()
 
 
-def _find_dunder(node_op, dunders) -> Union[Tuple[str, ...], None]:
+def _find_dunder(node_op, dunders) -> Union[tuple[str, ...], None]:
     dunder = None
     for op, candidate_dunder in dunders.items():
         if isinstance(node_op, op):
@@ -465,6 +518,78 @@ def _find_dunder(node_op, dunders) -> Union[Tuple[str, ...], None]:
     return dunder
 
 
+def get_policy(context: EvaluationContext) -> EvaluationPolicy:
+    policy = copy(EVALUATION_POLICIES[context.evaluation])
+
+    for key, value in context.policy_overrides.items():
+        if hasattr(policy, key):
+            setattr(policy, key, value)
+    return policy
+
+
+def _validate_policy_overrides(
+    policy_name: EvaluationPolicyName, policy_overrides: dict
+) -> bool:
+    policy = EVALUATION_POLICIES[policy_name]
+
+    all_good = True
+    for key, value in policy_overrides.items():
+        if not hasattr(policy, key):
+            warnings.warn(
+                f"Override {key!r} is not valid with {policy_name!r} evaluation policy"
+            )
+            all_good = False
+    return all_good
+
+
+def _handle_assign(node: ast.Assign, context: EvaluationContext):
+    value = eval_node(node.value, context)
+    transient_locals = context.transient_locals
+    for target in node.targets:
+        if isinstance(target, (ast.Tuple, ast.List)):
+            # Handle unpacking assignment
+            values = list(value)
+            targets = target.elts
+            starred = [i for i, t in enumerate(targets) if isinstance(t, ast.Starred)]
+
+            # Unified handling: treat no starred as starred at end
+            star_or_last_idx = starred[0] if starred else len(targets)
+
+            # Before starred
+            for i in range(star_or_last_idx):
+                transient_locals[targets[i].id] = values[i]
+
+            # Starred if exists
+            if starred:
+                end = len(values) - (len(targets) - star_or_last_idx - 1)
+                transient_locals[targets[star_or_last_idx].value.id] = values[
+                    star_or_last_idx:end
+                ]
+
+                # After starred
+                for i in range(star_or_last_idx + 1, len(targets)):
+                    transient_locals[targets[i].id] = values[
+                        len(values) - (len(targets) - i)
+                    ]
+        else:
+            transient_locals[target.id] = value
+    return None
+
+
+def _extract_args_and_kwargs(node: ast.Call, context: EvaluationContext):
+    args = [eval_node(arg, context) for arg in node.args]
+    kwargs = {
+        k: v
+        for kw in node.keywords
+        for k, v in (
+            {kw.arg: eval_node(kw.value, context)}
+            if kw.arg
+            else eval_node(kw.value, context)
+        ).items()
+    }
+    return args, kwargs
+
+
 def eval_node(node: Union[ast.AST, None], context: EvaluationContext):
     """Evaluate AST node in provided context.
 
@@ -492,11 +617,78 @@ def eval_node(node: Union[ast.AST, None], context: EvaluationContext):
     The purpose of this function is to guard against unwanted side-effects;
     it does not give guarantees on protection from malicious code execution.
     """
-    policy = EVALUATION_POLICIES[context.evaluation]
+    policy = get_policy(context)
+
     if node is None:
         return None
+    if isinstance(node, (ast.Interactive, ast.Module)):
+        result = None
+        for child_node in node.body:
+            result = eval_node(child_node, context)
+        return result
+    if isinstance(node, ast.FunctionDef):
+        # we ignore body and only extract the return type
+        is_property = False
+
+        for decorator_node in node.decorator_list:
+            try:
+                decorator = eval_node(decorator_node, context)
+            except NameError:
+                # if the decorator is not yet defined this is fine
+                # especialy because we don't handle imports yet
+                continue
+            if decorator is property:
+                is_property = True
+
+        return_type = eval_node(node.returns, context=context)
+
+        if is_property:
+            context.transient_locals[node.name] = _resolve_annotation(
+                return_type, context
+            )
+            return None
+
+        def dummy_function(*args, **kwargs):
+            pass
+
+        dummy_function.__annotations__["return"] = return_type
+        dummy_function.__name__ = node.name
+        dummy_function.__node__ = node
+        context.transient_locals[node.name] = dummy_function
+        return None
+    if isinstance(node, ast.ClassDef):
+        # TODO support class decorators?
+        class_locals = {}
+        class_context = context.replace(transient_locals=class_locals)
+        for child_node in node.body:
+            eval_node(child_node, class_context)
+        bases = tuple([eval_node(base, context) for base in node.bases])
+        dummy_class = type(node.name, bases, class_locals)
+        context.transient_locals[node.name] = dummy_class
+        return None
+    if isinstance(node, ast.Assign):
+        return _handle_assign(node, context)
+    if isinstance(node, ast.AnnAssign):
+        if not node.simple:
+            # for now only handle simple annotations
+            return None
+        context.transient_locals[node.target.id] = _resolve_annotation(
+            eval_node(node.annotation, context), context
+        )
+        return None
     if isinstance(node, ast.Expression):
         return eval_node(node.body, context)
+    if isinstance(node, ast.Expr):
+        return eval_node(node.value, context)
+    if isinstance(node, ast.Pass):
+        return None
+    if isinstance(node, ast.Import):
+        # TODO: populate transient_locals
+        return None
+    if isinstance(node, (ast.AugAssign, ast.Delete)):
+        return None
+    if isinstance(node, (ast.Global, ast.Nonlocal)):
+        return None
     if isinstance(node, ast.BinOp):
         left = eval_node(node.left, context)
         right = eval_node(node.right, context)
@@ -575,7 +767,12 @@ def eval_node(node: Union[ast.AST, None], context: EvaluationContext):
         dunders = _find_dunder(node.op, UNARY_OP_DUNDERS)
         if dunders:
             if policy.can_operate(dunders, value):
-                return getattr(value, dunders[0])()
+                try:
+                    return getattr(value, dunders[0])()
+                except AttributeError:
+                    raise TypeError(
+                        f"bad operand type for unary {node.op}: {type(value)}"
+                    )
             else:
                 raise GuardRejection(
                     f"Operation (`{dunders}`) for",
@@ -611,9 +808,9 @@ def eval_node(node: Union[ast.AST, None], context: EvaluationContext):
             return eval_node(node.orelse, context)
     if isinstance(node, ast.Call):
         func = eval_node(node.func, context)
-        if policy.can_call(func) and not node.keywords:
-            args = [eval_node(arg, context) for arg in node.args]
-            return func(*args)
+        if policy.can_call(func):
+            args, kwargs = _extract_args_and_kwargs(node, context)
+            return func(*args, **kwargs)
         if isclass(func):
             # this code path gets entered when calling class e.g. `MyClass()`
             # or `my_instance.__class__()` - in both cases `func` is `MyClass`.
@@ -632,7 +829,13 @@ def eval_node(node: Union[ast.AST, None], context: EvaluationContext):
             func,  # not joined to avoid calling `repr`
             f"not allowed in {context.evaluation} mode",
         )
-    raise ValueError("Unhandled node", ast.dump(node))
+    if isinstance(node, ast.Assert):
+        # message is always the second item, so if it is defined user would be completing
+        # on the message, not on the assertion test
+        if node.msg:
+            return eval_node(node.msg, context)
+        return eval_node(node.test, context)
+    return None
 
 
 def _eval_return_type(func: Callable, node: ast.Call, context: EvaluationContext):
@@ -648,25 +851,48 @@ def _eval_return_type(func: Callable, node: ast.Call, context: EvaluationContext
     # but resolved by signature call we know the return type
     not_empty = sig.return_annotation is not Signature.empty
     if not_empty:
-        return _resolve_annotation(sig.return_annotation, sig, func, node, context)
+        return _resolve_annotation(sig.return_annotation, context, sig, func, node)
     return NOT_EVALUATED
 
 
-def _resolve_annotation(
-    annotation,
-    sig: Signature,
-    func: Callable,
-    node: ast.Call,
+def _eval_annotation(
+    annotation: str,
     context: EvaluationContext,
 ):
-    """Resolve annotation created by user with `typing` module and custom objects."""
-    annotation = (
+    return (
         _eval_node_name(annotation, context)
         if isinstance(annotation, str)
         else annotation
     )
+
+
+class _GetItemDuck(dict):
+    """A dict subclass that always returns the factory instance and claims to have any item."""
+
+    def __init__(self, factory, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self._factory = factory
+
+    def __getitem__(self, key):
+        return self._factory()
+
+    def __contains__(self, key):
+        return True
+
+
+def _resolve_annotation(
+    annotation: object | str,
+    context: EvaluationContext,
+    sig: Signature | None = None,
+    func: Callable | None = None,
+    node: ast.Call | None = None,
+):
+    """Resolve annotation created by user with `typing` module and custom objects."""
+    if annotation is None:
+        return None
+    annotation = _eval_annotation(annotation, context)
     origin = get_origin(annotation)
-    if annotation is Self and hasattr(func, "__self__"):
+    if annotation is Self and func and hasattr(func, "__self__"):
         return func.__self__
     elif origin is Literal:
         type_args = get_args(annotation)
@@ -676,26 +902,65 @@ def _resolve_annotation(
         return ""
     elif annotation is AnyStr:
         index = None
-        for i, (key, value) in enumerate(sig.parameters.items()):
-            if value.annotation is AnyStr:
-                index = i
-                break
-        if index is not None and index < len(node.args):
-            return eval_node(node.args[index], context)
+        if func and hasattr(func, "__node__"):
+            def_node = func.__node__
+            for i, arg in enumerate(def_node.args.args):
+                if not arg.annotation:
+                    continue
+                annotation = _eval_annotation(arg.annotation.id, context)
+                if annotation is AnyStr:
+                    index = i
+                    break
+            is_bound_method = (
+                isinstance(func, MethodType) and getattr(func, "__self__") is not None
+            )
+            if index and is_bound_method:
+                index -= 1
+        elif sig:
+            for i, (key, value) in enumerate(sig.parameters.items()):
+                if value.annotation is AnyStr:
+                    index = i
+                    break
+        if index is None:
+            return None
+        if index < 0 or index >= len(node.args):
+            return None
+        return eval_node(node.args[index], context)
     elif origin is TypeGuard:
-        return bool()
+        return False
+    elif origin is set or origin is list:
+        # only one type argument allowed
+        attributes = [
+            attr
+            for attr in dir(
+                _resolve_annotation(get_args(annotation)[0], context, sig, func, node)
+            )
+        ]
+        duck = _Duck(attributes=dict.fromkeys(attributes))
+        return _Duck(
+            attributes=dict.fromkeys(dir(origin())),
+            # items are not strrictly needed for set
+            items=_GetItemDuck(lambda: duck),
+        )
+    elif origin is tuple:
+        # multiple type arguments
+        return tuple(
+            _resolve_annotation(arg, context, sig, func, node)
+            for arg in get_args(annotation)
+        )
     elif origin is Union:
+        # multiple type arguments
         attributes = [
             attr
             for type_arg in get_args(annotation)
-            for attr in dir(_resolve_annotation(type_arg, sig, func, node, context))
+            for attr in dir(_resolve_annotation(type_arg, context, sig, func, node))
         ]
         return _Duck(attributes=dict.fromkeys(attributes))
     elif is_typeddict(annotation):
         return _Duck(
             attributes=dict.fromkeys(dir(dict())),
             items={
-                k: _resolve_annotation(v, sig, func, node, context)
+                k: _resolve_annotation(v, context, sig, func, node)
                 for k, v in annotation.__annotations__.items()
             },
         )
@@ -703,17 +968,19 @@ def _resolve_annotation(
         return _Duck(attributes=dict.fromkeys(dir(annotation)))
     elif origin is Annotated:
         type_arg = get_args(annotation)[0]
-        return _resolve_annotation(type_arg, sig, func, node, context)
+        return _resolve_annotation(type_arg, context, sig, func, node)
     elif isinstance(annotation, NewType):
-        return _eval_or_create_duck(annotation.__supertype__, node, context)
+        return _eval_or_create_duck(annotation.__supertype__, context)
     elif isinstance(annotation, TypeAliasType):
-        return _eval_or_create_duck(annotation.__value__, node, context)
+        return _eval_or_create_duck(annotation.__value__, context)
     else:
-        return _eval_or_create_duck(annotation, node, context)
+        return _eval_or_create_duck(annotation, context)
 
 
 def _eval_node_name(node_id: str, context: EvaluationContext):
-    policy = EVALUATION_POLICIES[context.evaluation]
+    policy = get_policy(context)
+    if node_id in context.transient_locals:
+        return context.transient_locals[node_id]
     if policy.allow_locals_access and node_id in context.locals:
         return context.locals[node_id]
     if policy.allow_globals_access and node_id in context.globals:
@@ -721,6 +988,8 @@ def _eval_node_name(node_id: str, context: EvaluationContext):
     if policy.allow_builtins_access and hasattr(builtins, node_id):
         # note: do not use __builtins__, it is implementation detail of cPython
         return getattr(builtins, node_id)
+    if policy.allow_auto_import and context.auto_import:
+        return context.auto_import(node_id)
     if not policy.allow_globals_access and not policy.allow_locals_access:
         raise GuardRejection(
             f"Namespace access not allowed in {context.evaluation} mode"
@@ -729,12 +998,11 @@ def _eval_node_name(node_id: str, context: EvaluationContext):
         raise NameError(f"{node_id} not found in locals, globals, nor builtins")
 
 
-def _eval_or_create_duck(duck_type, node: ast.Call, context: EvaluationContext):
-    policy = EVALUATION_POLICIES[context.evaluation]
+def _eval_or_create_duck(duck_type, context: EvaluationContext):
+    policy = get_policy(context)
     # if allow-listed builtin is on type annotation, instantiate it
-    if policy.can_call(duck_type) and not node.keywords:
-        args = [eval_node(arg, context) for arg in node.args]
-        return duck_type(*args)
+    if policy.can_call(duck_type):
+        return duck_type()
     # if custom class is in type annotation, mock it
     return _create_duck_for_heap_type(duck_type)
 
@@ -764,12 +1032,14 @@ SUPPORTED_EXTERNAL_GETITEM = {
 }
 
 
-BUILTIN_GETITEM: Set[InstancesHaveGetItem] = {
+BUILTIN_GETITEM: set[InstancesHaveGetItem] = {
     dict,
     str,  # type: ignore[arg-type]
     bytes,  # type: ignore[arg-type]
     list,
     tuple,
+    type,  # for type annotations like list[str]
+    _Duck,
     collections.defaultdict,
     collections.deque,
     collections.OrderedDict,
@@ -792,7 +1062,7 @@ list_non_mutating_methods = ("copy", "index", "count")
 set_non_mutating_methods = set(dir(set)) & set(dir(frozenset))
 
 
-dict_keys: Type[collections.abc.KeysView] = type({}.keys())
+dict_keys: type[collections.abc.KeysView] = type({}.keys())
 
 NUMERICS = {int, float, complex}
 
@@ -813,6 +1083,8 @@ ALLOWED_CALLS = {
     *_list_methods(str),
     tuple,
     *_list_methods(tuple),
+    bool,
+    *_list_methods(bool),
     *NUMERICS,
     *[method for numeric_cls in NUMERICS for method in _list_methods(numeric_cls)],
     collections.deque,
@@ -833,7 +1105,7 @@ ALLOWED_CALLS = {
     collections.Counter.most_common,
 }
 
-BUILTIN_GETATTR: Set[MayHaveGetattr] = {
+BUILTIN_GETATTR: set[MayHaveGetattr] = {
     *BUILTIN_GETITEM,
     set,
     frozenset,
@@ -872,6 +1144,7 @@ EVALUATION_POLICIES = {
         allow_builtins_access=True,
         allow_locals_access=True,
         allow_globals_access=True,
+        allow_getitem_on_types=True,
         allowed_calls=ALLOWED_CALLS,
     ),
     "unsafe": EvaluationPolicy(
