@@ -445,7 +445,10 @@ class ExceptionMessagePreferenceTest(unittest.TestCase):
 
     def test_jsondecodeerror_message(self):
         cell = "import json;json.loads('{\"a\": }')"
-        expected = "JSONDecodeError: Expecting value: line 1 column 7 (char 6)"
+        if platform.python_implementation() == "PyPy":
+            expected = "JSONDecodeError: Unexpected '}': line 1 column 7 (char 6)"
+        else:
+            expected = "JSONDecodeError: Expecting value: line 1 column 7 (char 6)"
         ip.run_cell("%xmode plain")
         with tt.AssertPrints(expected):
             ip.run_cell(cell)
